#ifndef SHRIMPSCRIPT_H
#define SHRIMPSCRIPT_H

#include <map>
#include <vector>
#include <queue>
#include <stack>
#include <string>
#include <cstdio>
#include <cstdlib>
using namespace std;

class ShrimpSyntaxException
{
private:
	string error, detail;
	int line;
public:
	ShrimpSyntaxException(string err, string detail, int line=0 ) {
		this->error = err;
		this->detail = detail;
		this->line = line;
	}
	const string& getError() const {
		return error;
	}
	const string& getDetail() const {
		return detail;
	}
	int getLine()const {
		return line;
	}
};

class ShrimpRuntimeException
{
private:
	string error, detail, func;
	int line;
public:
	ShrimpRuntimeException(string err, string detail, string func, int line=0 ) {
		this->error = err;
		this->detail = detail;
		this->line = line;
		this->func = func;
	}
	const string& getError() const {
		return error;
	}
	const string& getDetail() const {
		return detail;
	}
	const string& getFunction() const {
		return func;
	}
	int getLine()const {
		return line;
	}
};


enum {
	TypeNone = 0,
	TypeInteger,
	TypeString,
	TypeFloat
};

struct ShrimpVariable {
	char		type;
	char		global;
	char		constant;
	short		index;
	int		integerValue;
	float		floatValue;
	string		stringValue;
	ShrimpVariable() {
		type = TypeNone;
		integerValue = 0;
		floatValue = 0.0;
		stringValue = "";
		global = constant = 0;
	}
	string&		getString() {
		char tmp[32];
		switch (type) {
		case TypeString:
			return stringValue;
		case TypeInteger:
			itoa(integerValue, tmp, 10);
			stringValue = tmp;
			return stringValue;
		case TypeFloat:
			sprintf(tmp, "%f", floatValue);
			stringValue = tmp;
			return stringValue;
		default:
			stringValue = "undefined";
			return stringValue;
		}
	}
	int		getInteger() {
		switch (type) {
		case TypeInteger:
			return integerValue;
		case TypeFloat:
			return (int)floatValue;
		case TypeString:
			return atoi(stringValue.c_str());
		}
	}
	float		getFloat() {
		switch (type) {
		case TypeFloat:
			return floatValue;
		case TypeInteger:
			return (float)integerValue;
		case TypeString:
			return atof(stringValue.c_str());
		}
	}
	void		setInteger(int i) {
		integerValue = i;
		type = TypeInteger;
	}
	void		setString(string s) {
		stringValue = s;
		type = TypeString;
	}
	void		setFloat(float f) {
		floatValue = f;
		type = TypeFloat;
	}
};


struct ShrimpToken {
	int		line;
	int		type;
	string		value;
};

enum {
	ShrimpFunction = 0,
	ShrimpAssignment,
	ShrimpCondition,
	ShrimpLoop,
	ShrimpScope,
	ShrimpVar,
	ShrimpConst,
	ShrimpCall,
	ShrimpEval,
	ShrimpExternal,
	ShrimpExit,
	ShrimpSelect
};

/* 语法 */
struct ShrimpSyntaxNode {
	int 			line;
	int			type;	//FUNCTION, VAR, IF, DO, CALL, ASSIGN, EVAL, CONST/
	string			symbolName;
	ShrimpSyntaxNode*	left;
	ShrimpSyntaxNode*	right;
	ShrimpSyntaxNode*	parent;
	ShrimpSyntaxNode*	function;
	int			operation;	//+, -, *, / or >=, <=, >, <, <>, =
	/* condition */
	ShrimpSyntaxNode*	expression;
	ShrimpSyntaxNode*	conditionTrue;
	ShrimpSyntaxNode*	conditionFalse;
	/* function */
	int			variableCount;
	map<string, ShrimpVariable*>	localVariables;
	/* scope */
	vector<ShrimpSyntaxNode*>	sentences;
	/* function call */
	vector<ShrimpSyntaxNode*>	parameters;
	/* variable */
	ShrimpVariable* 	value;
	ShrimpSyntaxNode(ShrimpSyntaxNode* parent) {
		this->parent = parent;
		if (parent)
			this->function = parent->function;
		else
			this->function = 0;
		this->line = 0;
		this->type = 0;
	}
	ShrimpSyntaxNode(ShrimpSyntaxNode* parent, ShrimpSyntaxNode* left, ShrimpSyntaxNode* right, int line, string name) {
		this->parent = parent;
		if (parent)
			this->function = parent->function;
		else
			this->function = 0;
		this->left = left;
		this->right = right;
		this->line = line;
		this->value = 0;
		this->symbolName = name;
	}
};

enum InterpretState {
	ISContinue = 0,
	ISExitSelect = 1,
	ISExitLoop = 2,
	ISExitFunction = 5,
	ISExitProgram = 9,
};

class ShrimpScript
{
private:
	queue<ShrimpToken>*		currentTokenSet;
	map<string, ShrimpSyntaxNode*>	functions;
	map<string, ShrimpVariable*>	globals;
	string				lastError;
	vector<ShrimpVariable>		runtimeStack;
private:
	void		setLastError(string err, string detail, int line = -1);
	void		addToken(int line, string token);
	ShrimpToken*	getToken();
	void		popToken();
	void		pushToken(ShrimpToken tok);
	bool		parseSyntax();
	void		parseDIM(ShrimpSyntaxNode* function);
	void		parseCONST(ShrimpSyntaxNode* function);
	void		parseFUNCTION();
	void		parseIMPORT();
	void		parseCALL(ShrimpSyntaxNode* p);
	void		parseBody(ShrimpSyntaxNode* p);
	void		endLine();
	ShrimpSyntaxNode*parseExpression(ShrimpSyntaxNode* p);
	ShrimpSyntaxNode*buildOperation(string op, stack<ShrimpSyntaxNode*>& varStack, int line);
	ShrimpVariable*	getVariable(ShrimpSyntaxNode* p, string name, bool create);
	int		getOperatorPriority(string op);
	void		invokeExternal(string name, int paraCount);
	ShrimpVariable	invokeFunction(string name, int paraCount);
	ShrimpVariable	call(ShrimpSyntaxNode* p);
	void		toUpperString(string& str);
	int		interpret(ShrimpSyntaxNode* p);
	ShrimpVariable& getRuntimeVar(ShrimpSyntaxNode* func, const string& name);
	ShrimpVariable& getRuntimeVar(ShrimpSyntaxNode* p);
	ShrimpVariable	evaluate(ShrimpSyntaxNode* p);
	void		remove(ShrimpSyntaxNode* p);
public:
	ShrimpScript();
	~ShrimpScript();
	string		getLastError() const;
	bool		load(string file);
	bool		parse(string code);
	bool		addFunction(string name, void (*func)(vector<ShrimpVariable>& stkVar, ShrimpVariable& result, int paraCount) );
	bool		run(string function);
};

#endif // SHRIMPSCRIPT_H
